/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#include <config.h>

#include <cups/cups.h>
#include <cups/http.h>
#include <cups/ipp.h>
#include <unistd.h>
#include <sys/types.h>
#include <glib.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtkradiobutton.h>
#include <gtk/gtktreeview.h>
#include <glade/glade.h>
#include <glib/gi18n.h>
#include <libgnomeui/gnome-druid.h>
#include <libgnomeui/gnome-ui-init.h>

#include <libgnomecups/gnome-cups-ui-init.h>
#include <libgnomecups/gnome-cups-request.h>
#include <libgnomecups/gnome-cups-util.h>
#include <libgnomecups/gnome-cups-ui-util.h>
#include <libgnomecups/gnome-cups-ui-driver.h>
#include <libgnomecups/gnome-cups-ui-connection.h>
#include <libgnomecups/gnome-cups-permission.h>

#include "druid-helper.h"

static GCupsDriverSelector *
driver_selector (GladeXML *xml)
{
	return (GCupsDriverSelector *) glade_xml_get_widget (xml, "driver_selector");
}
static GCupsConnectionSelector *
connection_selector (GladeXML *xml)
{
	return (GCupsConnectionSelector *) glade_xml_get_widget (xml, "connection_selector");
}

static void
set_selected_uri (GladeXML *xml, char const *uri)
{
	g_object_set_data_full (G_OBJECT (xml),
		"selected_uri", g_strdup (uri), g_free);
}

static char *
get_selected_uri (GladeXML *xml)
{
	char *str = g_object_get_data (G_OBJECT (xml), "selected_uri");
	g_return_val_if_fail (str != NULL, NULL);

	return g_strdup (str);
}

static void
cb_model_guess (G_GNUC_UNUSED GCupsConnectionSelector *cs,
		char const *model, GCupsDriverSelector *ds)
{
	gcups_driver_selector_set_nickname (ds, model);
}

static void
connection_page_setup (GladeXML *xml)
{
	GtkWidget *w = glade_xml_get_widget (xml, "connection_selector");
	g_signal_connect_swapped (w,
		"changed",
		G_CALLBACK (druid_update_sensitivities), xml);
	g_signal_connect_object (w,
		"model-guess",
		G_CALLBACK (cb_model_guess), driver_selector (xml), 0);
}

static void
connection_page_sensitivity (GladeXML *xml, gboolean *back, gboolean *next)
{
	char *uri = gcups_connection_selector_get_uri (connection_selector (xml));
	*next = (uri != NULL);
	g_free (uri);

	*back = FALSE;
}

static GtkWidget *
connection_page_next (GladeXML *xml)
{
	char *uri = gcups_connection_selector_get_uri (connection_selector (xml));
	GtkWidget *res = NULL;

	if (uri != NULL) {
		gcups_connection_selector_queue_guess (
			connection_selector (xml));
		res = glade_xml_get_widget (xml, "driver_page");
	}

	set_selected_uri (xml, uri);

	g_free (uri);
	return res;
}

/****************************************************************************/

static void
driver_page_setup (GladeXML *xml)
{
	g_signal_connect_swapped (glade_xml_get_widget (xml, "driver_selector"),
		"changed",
		G_CALLBACK (druid_update_sensitivities), xml);
}

static void
driver_page_sensitivity (GladeXML *xml, gboolean *back, gboolean *next)
{
	*back = TRUE;
	*next = NULL != gcups_driver_selector_get (driver_selector (xml));

	GCupsPPD const  *ppd  = gcups_driver_selector_get (driver_selector (xml));

	if (ppd && ppd->model) {
		g_signal_handlers_block_by_func(glade_xml_get_widget (xml, "printer-name-entry"), G_CALLBACK (druid_update_sensitivities), xml);
		GtkEntry *printer_name_entry = (GtkEntry*) glade_xml_get_widget(xml, "printer-name-entry");
		/* Clear invalid names (spaces + slashes) so that the default is always a valid one */
		char *name, *ptr;
		name = g_strdup(ppd->model);	
	
		for (ptr = name ; *ptr ; ptr++)
			if (*ptr == ' ' || *ptr == '/')
				*ptr = '-';

		gtk_entry_set_text(printer_name_entry, name);

		g_free(name);
		g_signal_handlers_unblock_by_func(glade_xml_get_widget (xml, "printer-name-entry"), G_CALLBACK (druid_update_sensitivities), xml);
		}
}


static GtkWidget *
driver_page_back (GladeXML *xml)
{
	return glade_xml_get_widget (xml, "connection_page");
}


// This function is just a 'fake' to avoid conflicts during compilation with debian patch properties_on_add.patch
static gboolean
add_cups_printer (GladeXML *xml, char const *device_uri, GCupsPPD  const *ppd, char const *printer_name)
{
	return FALSE;
}
// Note: The (debian/ubuntu) packager needs to fix this conflict by removing properties_on_add.patch
// and rename the function extended_add_cups_printer  to   add_cups_printer
// for questions feel free to contact me:    amos.brocco@unifr.ch

static gboolean
extended_add_cups_printer (GladeXML *xml, char const *device_uri, GCupsPPD  const *ppd, char const *printer_name, char const *printer_info, char const *printer_location)
{
	GError *err = NULL;
	char local_uri [HTTP_MAX_URI+1];
	ipp_t *request = gnome_cups_request_new (CUPS_ADD_PRINTER);
	g_snprintf (local_uri, sizeof local_uri - 1,
		    "ipp://localhost/printers/%s", printer_name);
	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
		      "printer-uri", NULL, local_uri);
	ippAddString (request,	IPP_TAG_PRINTER, IPP_TAG_NAME,
		      "printer-name", NULL, gnome_cups_strdup (printer_name));
	ippAddString (request,	IPP_TAG_PRINTER, IPP_TAG_TEXT,
		      "printer-info", NULL, printer_info);
	ippAddString (request,	IPP_TAG_PRINTER, IPP_TAG_TEXT,
		      "printer-location", NULL, printer_location);
	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_NAME,
		      "ppd-name", NULL, gnome_cups_strdup (ppd->filename));
	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_URI,
		      "device-uri", NULL, gnome_cups_strdup (device_uri));
	ippAddBoolean (request, IPP_TAG_PRINTER, "printer-is-accepting-jobs", 1);
	ippAddInteger(request, IPP_TAG_PRINTER, IPP_TAG_ENUM, "printer-state",
		      IPP_PRINTER_IDLE);
	ippDelete (gnome_cups_request_execute (request, NULL, "/admin/", &err));

	if (err != NULL) {
		gnome_cups_error_dialog ((GtkWindow *)glade_xml_get_widget (xml, "add_printer_window"),
					 _("Couldn't add printer"), err);
		g_error_free (err);
		return FALSE;
	}

	return TRUE;
}

static void
driver_page_prepare (GladeXML *xml)
{
#if 0
	/* Blah.  GnomeDruid is stupid.  finish and next are distinct buttons internally
	 * and there is no way to desensitize finish.  We'll need to cheat */
	GtkWidget *druid = glade_xml_get_widget (xml, "add_printer_druid");
	gnome_druid_set_show_finish (GNOME_DRUID (druid), TRUE);
#endif
}

static GtkWidget *
driver_page_next (GladeXML *xml)
{
	char  *name, *uri, *ptr;
	GCupsPPD const  *ppd  = gcups_driver_selector_get (driver_selector (xml));
	GtkWidget *res = NULL;
	
	if (ppd == NULL)
		return NULL;

	uri  = get_selected_uri (xml);

	if (uri != NULL) {
		res = glade_xml_get_widget (xml, "printer_info_page");
	}

	g_free (uri);
	return res;
}

/****************************************************************************/

static void
printer_info_page_setup (GladeXML *xml)
{
	GtkWidget *w = glade_xml_get_widget (xml, "printer-name-entry");
	g_signal_connect_swapped (w,
		"changed",
		G_CALLBACK (druid_update_sensitivities), xml);
}

static gboolean 
check_printer_name(GladeXML *xml) 
{
	/* Checks whether the name is correctly filled */
	GtkEntry *printer_name_entry = (GtkEntry *)glade_xml_get_widget (xml, "printer-name-entry");
	GtkLabel *warnings_label = (GtkLabel *)glade_xml_get_widget (xml, "warnings-label");

	char  *printer_name, *ptr;
	gboolean has_invalid_chars = FALSE;
	unsigned i = 0;

	printer_name = g_strchomp(g_strdup(gtk_entry_get_text(printer_name_entry)));

	if(!strlen(printer_name)) {
		gtk_label_set_markup(warnings_label,_("<i>Printer name cannot be empty.</i>"));           
		g_free (printer_name);              
		return FALSE;
		}

	/* Checks that the name contains no spaces or slashes (cf. Ubuntu Bug #32085) */
	for (ptr = printer_name ; *ptr ; ptr++)
		if (*ptr == ' ' || *ptr == '/') {
			has_invalid_chars = TRUE;
			break;
			}

	
	/* If the name is not valid show a warning message */
	if(has_invalid_chars) {
		gtk_label_set_markup(warnings_label,_("<i>Printer name contains invalid characters or spaces.</i>"));           
		g_free (printer_name);              
		return FALSE;
		}
	
	
	g_free (printer_name);  
	gtk_label_set_markup(warnings_label,"");
	return TRUE;
}

static void
printer_info_page_sensitivity (GladeXML *xml, gboolean *back, gboolean *next)
{
	GtkWidget *druid = glade_xml_get_widget (xml, "add_printer_druid");
	*back = TRUE;
	*next = FALSE != check_printer_name (xml);

	/* see driver_page_prepare for an explanation of the druid hack */
	gnome_druid_set_show_finish (GNOME_DRUID (druid), *next);
}

static GtkWidget *
printer_info_page_back (GladeXML *xml)
{
	return glade_xml_get_widget (xml, "driver_page");
}


static void
printer_info_page_prepare (GladeXML *xml)
{
	/* The printer name defaults to the printer model */	
	GtkEntry *printer_name_entry = (GtkEntry*) glade_xml_get_widget(xml, "printer-name-entry");
	GCupsPPD const  *ppd  = gcups_driver_selector_get (driver_selector (xml));
	GtkLabel *warnings_label = (GtkLabel *)glade_xml_get_widget (xml, "warnings-label");
	char  *name;

	gtk_misc_set_alignment(GTK_MISC(warnings_label), 0.0f, 0.0f);
	
	name = g_strdup(gtk_entry_get_text(printer_name_entry));
	
	if(!strlen(name))
		gtk_entry_set_text(printer_name_entry, ppd->model);

	g_free(name);
}

static GtkWidget *
printer_info_page_next (GladeXML *xml)
{
	GCupsPPD const  *ppd  = gcups_driver_selector_get (driver_selector (xml));
	char  *original_name, *name, *uri, *description, *location;
	GtkEntry *printer_name_entry = (GtkEntry*) glade_xml_get_widget(xml, "printer-name-entry");
	GtkEntry *printer_description_entry = (GtkEntry*) glade_xml_get_widget(xml, "printer-description-entry");
	GtkEntry *printer_location_entry = (GtkEntry*) glade_xml_get_widget(xml, "printer-location-entry");
	GtkWidget *parent_window = glade_xml_get_widget (xml, "add_printer_window");
	GList *existing;
	unsigned i = 0;
	gboolean name_changed = FALSE;

	uri  = get_selected_uri (xml);
	name = g_strchomp(g_strdup(gtk_entry_get_text(printer_name_entry)));
	description = g_strchomp(g_strdup(gtk_entry_get_text(printer_description_entry)));
	location = g_strchomp(g_strdup(gtk_entry_get_text(printer_location_entry)));
	original_name = g_strchomp(g_strdup(gtk_entry_get_text(printer_name_entry)));

	existing = gnome_cups_get_printers ();

	while (NULL != g_list_find_custom (existing, name, (GCompareFunc)strcasecmp )) {
		g_free (name);
		name = g_strdup_printf ("%s-%d", original_name, ++i);
		name_changed = TRUE;
		} 

	g_list_foreach (existing, (GFunc)g_free, NULL);
	g_list_free (existing);

	if (add_cups_printer (xml, uri, ppd, name)) {
		g_object_unref (xml);
		gtk_main_quit ();
	} // The above code (in the if) will never be executed: leaved to make a patch happy (see explication above)
 
	if(name_changed) {
	 	GtkWidget *dialog;
 
                dialog = gtk_message_dialog_new
                        (GTK_WINDOW (parent_window),
                         GTK_DIALOG_DESTROY_WITH_PARENT,
                         GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
                         _("Printer '%s' is already installed on this system. This printer has been renamed to '%s'."),
			 original_name,
                         name);
		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
		}

	if (extended_add_cups_printer (xml, uri, ppd, name, description, location)) {
		g_object_unref (xml);
		gtk_main_quit ();
	} 

	g_free (uri);
	g_free (name);
	g_free (description);
	g_free (location);
	g_free (original_name);
	return NULL;
}


/****************************************************************************/

static DruidPageDescription pages[] = {
	{
		"connection_page",
		connection_page_setup,
		NULL,
		connection_page_sensitivity,
		NULL,
		connection_page_next,
	},
	{
		"driver_page",
		driver_page_setup,
		driver_page_prepare,
		driver_page_sensitivity,
		driver_page_back,
		driver_page_next
	},
	{
		"printer_info_page",
		printer_info_page_setup,
		printer_info_page_prepare,
		printer_info_page_sensitivity,
		printer_info_page_back,
		printer_info_page_next
	},
	{ NULL }
};

static int
delete_event_cb (GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
	gtk_main_quit ();
	return FALSE;
}

static void
cancel_cb (GtkWidget *widget, gpointer user_data)
{
	gtk_main_quit ();
}

GtkWindow *progress_window = NULL;
static GtkWindow*
create_progress_window ()
{
       GtkWindow *window;
       GtkWidget *vbox, *label;
       //GtkProgressBar *progress;
       guint sid;

       window = GTK_WINDOW (gtk_window_new (GTK_WINDOW_TOPLEVEL));
       //gtk_window_set_title (GTK_WINDOW (window), _("Add a Printer"));
       gtk_window_set_decorated(window, gtk_false);
       gtk_container_set_border_width (GTK_CONTAINER (window), 6);
       gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER_ALWAYS);
       g_signal_connect(window, "delete_event", G_CALLBACK(gtk_true), NULL);

       vbox = gtk_vbox_new (FALSE, 6);
       gtk_container_add (GTK_CONTAINER (window), vbox);

       label = gtk_label_new (g_strdup_printf (_("<b>Reading printer database ...</b>")));
       gtk_label_set_use_markup (GTK_LABEL(label), TRUE);
       gtk_misc_set_alignment (GTK_MISC(label), 0, 0.5);
       gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

       //progress = GTK_PROGRESS_BAR( gtk_progress_bar_new());
       //gtk_widget_show ( GTK_WIDGET (progress));
       //gtk_progress_bar_set_text (GTK_PROGRESS_BAR (progress), " ");
       //gtk_progress_bar_set_pulse_step (progress, 0.2);
       //gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (progress), FALSE, FALSE, 6);

       //sid = g_timeout_add (200, (GtkFunction) progress_timer, progress);
       //g_object_set_data(G_OBJECT(window), "pgb_id", GINT_TO_POINTER(sid));

       gtk_widget_show_all ( GTK_WIDGET (window));

       gdk_window_set_cursor(GTK_WIDGET(window)->window, gdk_cursor_new(GDK_WATCH));
       gtk_widget_grab_focus ( GTK_WIDGET (window));

       return window;
}

static gpointer
add_printer_main(gpointer data)
{
	GladeXML  *xml;
	GtkWidget *window;

	g_idle_remove_by_data(data);

	xml = glade_xml_new (GNOME_CUPS_MANAGER_DATADIR "/gnome-cups-add.glade",
			     "add_printer_window",
			     GETTEXT_PACKAGE);

	window = glade_xml_get_widget (xml, "add_printer_window");
	set_window_icon (window, "gnome-dev-printer-new");
	g_signal_connect (window,
		"delete_event",
		G_CALLBACK (delete_event_cb), NULL);
	g_signal_connect (glade_xml_get_widget (xml, "add_printer_druid"),
		"cancel",
		G_CALLBACK (cancel_cb), NULL);

	druid_pages_setup (xml, pages);
	if (progress_window) {
		gdk_window_set_cursor(GTK_WIDGET(progress_window)->window, gdk_cursor_new(GDK_LAST_CURSOR));
		//g_source_remove(GPOINTER_TO_INT(g_object_get_data(G_OBJECT(progress_window), "pgb_id")));
		gtk_widget_destroy (GTK_WIDGET(progress_window));
	}

	gtk_widget_show (window);

	return NULL; 
}

int
main (int argc, char *argv[])
{
	gnome_program_init ("gnome-cups-add",
			    VERSION,
			    LIBGNOMEUI_MODULE, argc, argv,
			    GNOME_PROGRAM_STANDARD_PROPERTIES,
			    GNOME_PARAM_HUMAN_READABLE_NAME, _("Add a Printer"),
			    NULL);
	glade_init ();
	gnome_cups_ui_init (argv[0]);

	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	if (!gnome_cups_can_admin ()) {
		if (gnome_cups_spawn ("gnome-cups-add", argc-1,
				      (char const **)(argv+1), TRUE, NULL))
			exit (0);
		else
			exit (1);
	}

	progress_window = create_progress_window();
	g_timeout_add(300, add_printer_main, progress_window);

	gtk_main ();

	return 0;
}
